//Ethan Alexander Shulman 2024

var cameraX = 1024, cameraY = 1024, aspectX = 1, aspectY = 1, cameraZoom = 256,
winWidth = 0, winHeight = 0,
pointShader, pixelShader,
quadIndexBuffer,

SPRITE_ATTRIBS,
SPRITE_INDICES_LUT = [2,1,0, 3,1,2];


//initialize rendering systems
function InitializeRendering() {
	SPRITE_ATTRIBS = [new MeshAttrib("vPos",4,4,gl.FLOAT)];
	
	//create one big index buffer of indices for sprite quads
	var ind = new Uint16Array(16384*6), i = 0, fi = 0;
	for (var q = 0; q < 16384; q++) {
		ind[i++] = fi+2;
		ind[i++] = fi+1;
		ind[i++] = fi;
		ind[i++] = fi+3;
		ind[i++] = fi+1;
		ind[i++] = fi+2;
		fi += 4;
	}
	quadIndexBuffer = gl.createBuffer();
	gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, quadIndexBuffer);
	gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, ind, gl.STATIC_DRAW);
	
	//load shaders
	var svert = LoadShader("SpriteVert.glsl", gl.VERTEX_SHADER),
		pointFrag = LoadShader("Point.glsl", gl.FRAGMENT_SHADER),
		pixelFrag = LoadShader("Pixel.glsl", gl.FRAGMENT_SHADER);
	pointShader = new Shader(svert,pointFrag,SPRITE_ATTRIBS);
	pixelShader = new Shader(svert,pixelFrag,SPRITE_ATTRIBS);
	gl.deleteShader(svert);
	gl.deleteShader(pointFrag);
	gl.deleteShader(pixelFrag);

	//window display resizing
	OnWindowResize();
	window.addEventListener("resize", OnWindowResize);
}


//on window resize event
function OnWindowResize() {
	SetDisplaySize(winWidth = window.innerWidth, winHeight = window.innerHeight, true);
	
	//resize aspect
	var len = 1/(display.width+display.height),
		asX = display.width*len,
		asY = display.height*len;
	len = 1/Math.sqrt(asX*asY);
	aspectX = asX*len;
	aspectY = asY*len;
}


//world and screen coordinate conversion
function WorldToScreen(wx,wy) {
	returnY = ((wy-cameraY)/(aspectY*cameraZoom)*-0.5+0.5)*display.height;
	return ((wx-cameraX)/(aspectX*cameraZoom)*0.5+0.5)*display.width;
}
function ScreenToWorld(sx,sy) {
	returnY = (sy/display.height*-2+1)*(aspectY*cameraZoom) + cameraY;
	return (sx/display.width*2-1)*(aspectX*cameraZoom) + cameraX;
}
function ScaleWorldToScreen(wsx,wsy) {
	returnY = wsy/(aspectY*cameraZoom)*display.height*-0.5;
	return wsx/(aspectX*cameraZoom)*display.width*0.5;
}
function ScaleScreenToWorld(ssx,ssy) {
	returnY = (ssy/display.height)*(aspectY*cameraZoom)*-2;
	return (ssx/display.width)*(aspectX*cameraZoom)*2;
}



//sprite batch, multiple sprites sharing the same shader/uniforms rendered together
/**@constructor*/
function SpriteBatch() {
	this.meshes = [];
	this.bufferSize = 0;
	this.renderCount = 0;
	this.verts = null;
	this.indices = null;
	this.lines = false;
}

//update sprite batch
SpriteBatch.prototype.Update = function(sps, staticDraw) {
	//resize buffers
	var nsprites = sps.length, bufSz = this.bufferSize,
		reqMeshes = Math.ceil(nsprites/16384), reqBufSz = (reqMeshes > 1 ? 16384 : nsprites),
		allocMeshes = (this.meshes.length < reqMeshes),
		allocBuf = (bufSz < reqBufSz);
	if (allocMeshes || allocBuf) {
		if (allocBuf) {
			if (bufSz === 0) bufSz = 1;
			while (bufSz < reqBufSz) {
				bufSz *= 2;
			}
			this.verts = new Float32Array(16*bufSz);
			this.bufferSize = bufSz;
		}
		if (allocMeshes) {
			for (var i = this.meshes.length; i < reqMeshes; i++) {
				var msh = new Mesh(SPRITE_ATTRIBS);
				msh.indexBuffer = quadIndexBuffer;
				msh.sharedIndices = true;
				this.meshes.push(msh);
			}
		}
	}
	
	//build mesh vertices
	var lines = this.lines,
		va = this.verts, vi = 0,
		meshes = this.meshes, mi = 0,
		s = 0;
	while (s < nsprites) {
		var iter = Math.min(16384,nsprites-s);
		vi = 0;
		for (var i = 0; i < iter; i++) {
			var sprite = sps[s],
				x = sprite.x, y = sprite.y,
				sx = sprite.scaleX, sy = sprite.scaleY;
			if (lines) {
				var acs = Math.cos(sprite.angle), asn = Math.sin(sprite.angle);
				
				//top left
				va[vi++] = x - sx*acs + sy*asn;
				va[vi++] = y - sx*-asn + sy*acs;
				va[vi++] = -1;
				va[vi++] = 1;
			
				//top right
				va[vi++] = x + sx*acs + sy*asn;
				va[vi++] = y + sx*-asn + sy*acs;
				va[vi++] = 1;
				va[vi++] = 1;
			
				//bottom left
				va[vi++] = x - sx*acs - sy*asn;
				va[vi++] = y - sx*-asn - sy*acs;
				va[vi++] = -1;
				va[vi++] = -1;

				//bottom right
				va[vi++] = x + sx*acs - sy*asn;
				va[vi++] = y + sx*-asn - sy*acs;
				va[vi++] = 1;
				va[vi++] = -1;
			
			} else {
				//top left
				va[vi++] = x - sx;
				va[vi++] = y + sy;
				va[vi++] = -1;
				va[vi++] = 1;
			
				//top right
				va[vi++] = x + sx;
				va[vi++] = y + sy;
				va[vi++] = 1;
				va[vi++] = 1;
			
				//bottom left
				va[vi++] = x - sx;
				va[vi++] = y - sy;
				va[vi++] = -1;
				va[vi++] = -1;

				//bottom right
				va[vi++] = x + sx;
				va[vi++] = y - sy;
				va[vi++] = 1;
				va[vi++] = -1;
			}
			s++;
		}
		meshes[mi++].Set((vi === va.length ? va : va.subarray(0,vi)), staticDraw ? gl.STATIC_DRAW : gl.DYNAMIC_DRAW);
	}
	this.renderCount = nsprites;
}

//render sprite batch meshes
SpriteBatch.prototype.Render = function() {
	var rcount = this.renderCount, mi = 0,
		meshes = this.meshes;
	while (rcount) {
		var iter = Math.min(16384,rcount);
		meshes[mi++].Bind();
		gl.drawElements(gl.TRIANGLES,iter*6,gl.UNSIGNED_SHORT,0);
		rcount -= iter;
	}
}


//sprite rendered via sprite batch
/**@constructor*/
function Sprite() {
	this.x = 0;
	this.y = 0;
	this.scaleX = 1;
	this.scaleY = 1;
	this.angle = 0;
}


//render
function Render() {
	var camSX = 1/(aspectX*cameraZoom),
		camSY = 1/(aspectY*cameraZoom);
	display.Bind();
	gl.clearColor(0,0,0,1);
	gl.clear(gl.COLOR_BUFFER_BIT);

	pixelShader.Bind();
	gl.uniform4f(pixelShader["camera"], cameraX,cameraY,camSX,camSY);
	for (var fi = 0; fi < activeFactions.length; fi++) {
		var pxFaction = activeFactions[fi];
		gl.uniform4f(pixelShader["color"], pxFaction.red,pxFaction.green,pxFaction.blue,1);
		pxFaction.pixelsBatch.Update(pxFaction.pixels, false);
		pxFaction.pixelsBatch.Render();
	}
}
